/*
 *  txt2table
 *  Copyright(C) 1999 by C-yan
 *
 *  $Log: txt2csv.c,v $
 *  Revision 2.6  1999/05/19 13:08:01  anonymous
 *  t@CTCYl printf Xy[X폜B
 *
 *  Revision 2.5  1999/05/19 00:31:42  anonymous
 *  TeX`o͂ǉB
 *
 *  Revision 2.4  1999/05/13 09:36:49  anonymous
 *  htmlo͎'"'̃GXP[v̑gݍ݁B ̑B
 *
 *  Revision 2.3  1999/05/05 16:49:46  anonymous
 *  \[X̌ȂB
 *
 *  Revision 2.2  1999/05/05 12:33:49  anonymous
 *  ex_html[hǉB
 *
 *  Revision 2.0  1999/05/04 23:19:05  anonymous
 *  \[X̌BiRg̒ǉj
 *
 *  Revision 1.9  1999/05/04 20:13:30  anonymous
 *  R}hC̃IvVɂďo͂肩悤ɂB
 *
 *  Revision 1.8  1999/05/04 18:01:53  anonymous
 *  YẴGXP[vǉB(char_escape֐j
 *
 *  Revision 1.7  1999/05/04 16:32:20  anonymous
 *  txt2html @\ǉBi܂\[Xxł؂肩Ȃj
 *  ̑낢ƒۉB
 *
 *  Revision 1.6  1999/05/03 22:54:59  anonymous
 *  s̐擪̃XLbvs悤ɏCB
 *  eLXgɃRg悤ɕύXB
 *
 *  Revision 1.5  1999/05/03 20:39:24  anonymous
 *  \̌đ啝ȏB
 *
 *  Revision 1.4  1999/05/03 17:19:59  anonymous
 *  put_item }N
 *
 *  Revision 1.3  1999/05/03 16:21:11  anonymous
 *  assign_file ǉ
 *
 *  Revision 1.2  1999/05/02 20:22:28  anonymous
 *  ŊB
 *
 */

#include <stdio.h>
#include <string.h>

enum {csv, html, ex_html, tex};

void operate_commandline(int argc, char **argv, char *infile, char *outfile, int *kind);
void print_usage(void);
void print_version(void);

int txt2table(int kind, char *infile, char *outfile);
FILE *assign_file(char *filename, char *mode, FILE *default_fp);

int skip_first_space(FILE *infp);

void file_start(int kind, FILE *outfp);
void line_start(int kind, FILE *outfp);

int get_item(int kind, char *item, FILE *infp);
int read_item(char *data, FILE *infp);
int char_escape(int kind, char *data);
void html_escape(char **data_ptr, char **tmp_ptr);
void tex_escape(char **data_ptr, char **tmp_ptr);
int get_format(int kind, char *format, int state, char *data);

#define put_item(a, b) fputs(a, b)	

void line_end(int kind, FILE *outfp);
void file_end(int kind, FILE *outfp);

int skip_space(FILE *infp);
int skip_line(FILE *infp);

int main(int argc, char **argv)
{
	int kind;
	char infile[FILENAME_MAX], outfile[FILENAME_MAX];
	
	*infile = *outfile = '\0';
	kind = csv;

	operate_commandline(argc, argv, infile, outfile, &kind);

	txt2table(kind, infile, outfile);

	return 0;
}

void operate_commandline(int argc, char **argv, char *infile, char *outfile, int *kind)
{
	int i;
	char *c;
	
	for(i = 2; i <= argc; i++){
		c = argv[i - 1];
		
		if(     strcmp(c, "-csv"     ) == 0)
			*kind = csv;
		else if(strcmp(c, "-html"    ) == 0)
			*kind = html;
		else if(strcmp(c, "-ex_html" ) == 0)
			*kind = ex_html;
		else if(strcmp(c, "-tex"     ) == 0)
			*kind = tex;
		else if(strcmp(c, "-i"       ) == 0)
			strcpy(infile,  argv[i++]);
		else if(strcmp(c, "-o"       ) == 0)
			strcpy(outfile, argv[i++]);
		else if(strcmp(c, "--help"   ) == 0){
			print_usage();
			exit(0);
		}
		else if(strcmp(c, "--version") == 0){
			print_version();
			exit(0);
		}
	}
}

void print_usage(void)
{
	printf(
		"Usage:"                                                              "\n"
		"  txt2table [options]"                                               "\n"
		"Options:"                                                            "\n"
		"  -csv         Output as csv format."                                "\n"
		"  -html        Output as html format."                               "\n"
		"  -ex_html     Output as html format but don't escape \", <, >, &."  "\n"
		"  -tex         Output as TeX format."                                "\n"
		"  -i filename  Name input file."                                     "\n"
		"  -o filename  Name output file."                                    "\n"
		"  --help       Print this help message."                             "\n"
		"  --version    Print version information."                           "\n"
	);
}

void print_version(void)
{
	printf(
		"txt2table v1.61 #11"                  "\n"
		"Copyright(C)1999 by C-yan"            "\n"
		"1999/05/19 Wed. (Since 1999/05/03)"   "\n"
	);
}

#define ITEM_SIZE_MAX 1000
int txt2table(int kind, char *infile, char *outfile)
{
	int state;
	FILE *infp, *outfp;
	char item[ITEM_SIZE_MAX];

	infp  = assign_file(infile,  "r",  stdin );
	outfp = assign_file(outfile, "w+", stdout);
	
	state = 0;
	file_start(kind, outfp);
	while(state != EOF){
		state = skip_first_space(infp);
		
		if(state != EOF){
			line_start(kind, outfp);
			while((state != '\n') && (state != EOF)){
				state = get_item(kind, item, infp);
				put_item(item, outfp);
			}
			line_end(kind, outfp);
		}
	} 
	file_end(kind, outfp);

	return 0;
}

FILE *assign_file(char *filename, char *mode, FILE *default_fp)
{
	FILE *fp;
	
	if (*filename == '\0')
		fp = default_fp;
	else if ((fp = fopen(filename, mode)) == NULL){
		fprintf(stderr, "error: can't open %s", filename);
		exit(1);
	}
	
	return fp;
}

/* sRg΂es̐擪̃f[^̓o */
int skip_first_space(FILE *infp)
{
	int state;

	do{
		state = skip_space(infp);
		
		if(state == ';')                 /* ';' Ŏn܂s̓Rgs */
			state = skip_line(infp);

	}while(state == '\n');
	                                  
	if(state != EOF)                     /* s΂ۂɐǂ݂Ă܂ */
		ungetc(state, infp);             /* Ԃ(EOF͉ԂȂ) */
	
	return state;
}

void line_start(int kind, FILE *outfp)
{
	if((kind == html) || (kind == ex_html))
		put_item("<TR>", outfp);
}

void line_end(int kind, FILE *outfp)
{
	if((kind == html) || (kind == ex_html))
		put_item("</TR>\n", outfp);
	else if(kind == csv)
		put_item("\n", outfp);
	else if(kind == tex)
		put_item(" \\\\ \\hline \n", outfp);
}

void file_start(int kind, FILE *outfp)
{
	if((kind == html) || (kind == ex_html))
		put_item("<TABLE BORDER=\"1\">\n<TBODY>\n", outfp);
	else if(kind == tex)
		put_item("\\begin{tabular}{} \\hline \n", outfp);
}

void file_end(int kind, FILE *outfp)
{
	if((kind == html) || (kind == ex_html))
		put_item("</TBODY>\n</TABLE>", outfp);
	else if(kind == tex)
		put_item("\\end{tabular}\n", outfp);
}

int read_item(char *data, FILE *infp)
{
	int state, prev;
	char *ptr;
	
	ptr = data;
	prev = state = 0;
	while(1){ 
		state = getc(infp);                   

		if((state == '\n') || (state == EOF)) /* sAt@C̏IA              */
			break;                            /* QAȏ̋󔒂f[^̐؂    */

		if((prev == ' ') && (state == ' ')){
			ptr--;                            /* ʂȃXy[XPf[^ɏo͂̂ */
			break;                            /* P߂                             */
		}

		*(ptr++) = state;
		prev  = state;
	}
	*ptr = '\0';
	
	return state;
}

#define FORMAT_SIZE 10
int get_item(int kind, char *item, FILE *infp)
{
	int state;
	char data[ITEM_SIZE_MAX], format[FORMAT_SIZE];
	
	state = read_item(data, infp);
	
	if(state == ' '){                           /* ̃f[^̓o */
		state = skip_space(infp); 
		if((state != '\n') && (state != EOF))
			ungetc(state, infp);
	}
	
	get_format(kind, format, state, data);
	
	sprintf(item, format, data);

	return state;
}

/* o͂̏(sprintfp) */
int get_format(int kind, char *format, int state, char *data) 
{
	*format = '\0';
	switch(kind){
		case csv:
			if(char_escape(kind, data) == 1)      /* f[^ ','   '"'  ' ' ꍇ */
				strcpy(format, "\"%s\"");         /* "" ł */
			else
				strcpy(format, "%s");
			
			if((state != '\n') && (state != EOF)) /* ̃f[^ݍsɂꍇɂ */
				strcat(format, ",");              /* ؂L */
			
			break;
		case html:
			char_escape(kind, data);              /* not break!! */
		case ex_html:
			strcat(format, "<TD>%s</TD>");
			break;
		case tex:
			char_escape(kind, data);
			strcpy(format, "%s");
			if((state != '\n') && (state != EOF)) /* ̃f[^ݍsɂꍇɂ */
				strcat(format, " & ");              /* ؂L */
			break;
	}

	return 0;
}

/* GXP[vȂƂȂGXP[v */
int char_escape(int kind, char *data)
{
	int quote; /* 񒆂 ','  '"'  ' ' ̂ǂꂩێ(csvp) */
	char *tmp_ptr, *data_ptr, tmp[ITEM_SIZE_MAX];
	
	quote = 0;
	tmp_ptr = tmp;
	switch(kind){
		case csv:
			for(data_ptr = data; *data_ptr != '\0'; data_ptr++){
				if((*data_ptr == '"') || (*data_ptr == ',') || (*data_ptr == ' '))
					quote = 1;

				if(*data_ptr == '"')      /* "  "" ɒu */
					*(tmp_ptr++) = '"';
				*(tmp_ptr++) = *data_ptr;
			}
			break;
		case html:
			for(data_ptr = data; *data_ptr != '\0'; data_ptr++)
				html_escape(&data_ptr, &tmp_ptr);
			break;
		case tex:
			for(data_ptr = data; *data_ptr != '\0'; data_ptr++)
				tex_escape(&data_ptr, &tmp_ptr);
			break;
	}
	*tmp_ptr = '\0';
	strcpy(data, tmp);
	return quote;
}

#define ESCAPESIZE_MAX 10
void html_escape(char **data_ptr, char **tmp_ptr)
{
	char s[ESCAPESIZE_MAX];
	
	switch(**data_ptr){
		case '<':
			strcpy(s, "&lt;");
			break;
		case '>':
			strcpy(s, "&gt;");
			break;
		case '&':
			strcpy(s, "&amp;");
			break;
		case '"':
			strcpy(s, "&quot;");
			break;
		default:
			sprintf(s, "%c", **data_ptr);
			break;
	}

	strcpy(*tmp_ptr, s);
	*tmp_ptr += strlen(s);
}

void tex_escape(char **data_ptr, char **tmp_ptr)
{
	char s[ESCAPESIZE_MAX];
	
	switch(**data_ptr){
		case '#': case '$': case '%': case '&': 
		case '_': case '{':	case '}':
			sprintf(s, "\\%c", **data_ptr);
			break;
		case '<': case '>': case '\\': case '^':
		case '|': case '~':
			sprintf(s, "\\verb@%c@", **data_ptr);
			break;
		default:
			*s     = **data_ptr;
			*(s+1) = '\0';
			break;
	}

	strcpy(*tmp_ptr, s);
	*tmp_ptr += strlen(s);
}

int skip_space(FILE *infp)
{
	int state;
	
	while((state = getc(infp)) == ' ')
		;
		
	return state;
}

int skip_line(FILE *infp)
{
	int state = 0;

	while((state != EOF) && (state != '\n'))
		state = getc(infp);
		
	return state;
}
